import os
import signal
import threading
import traceback
import subprocess

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType

from libs import logs
from libs.config import read_config
from libs.message import JupyterServerCommand, Message, WebappEndpoint
from project_manager import files, projects
from project_manager.interfaces import WorkspaceMetadata
from project_manager.interfaces import FileManagerMessageParams

log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space, workspace_info: WorkspaceMetadata, jupyter_server_config):
        super(MessageHandler, self).__init__(p2n_queue, user_space)
        self.jupyter_server_config = jupyter_server_config
        run_string = ["jupyter", "server", "--ip=0.0.0.0",
                      f"--port={jupyter_server_config['port']}",
                      f"--ServerApp.token={jupyter_server_config['token']}",
                      "--ServerApp.allow_origin=*"]
        log.info("Run: %s" % run_string)
        self.jupyter_server = subprocess.Popen(
            run_string, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    def jupyter_stdout_handler(self):
        while True:
            line = self.jupyter_server.stdout.readline()
            if not line:
                continue
            log.info(line.rstrip())

    def handle_message(self, message):
        try:
            if message.command_name == JupyterServerCommand.get_config:
                self._get_jupyter_server_config()
        except:
            pass

    def _get_jupyter_server_config(self):
        jupyter_server_config_message = Message(**{"webapp_endpoint": WebappEndpoint.Terminal, "command_name": JupyterServerCommand.get_config,
                                                   "content": self.jupyter_server_config, "error": False})
        self._send_to_node(jupyter_server_config_message)

    def shutdown(self):
        try:
            log.info('Shutting down pid=%s' % self.jupyter_server.pid)
            os.killpg(os.getpgid(self.jupyter_server.pid), signal.SIGTERM)
        except ProcessLookupError as error:
            # Sometime the the jupyter server might be exited before shutdown
            # so we have to catch this except to make sure the program will exit gracefully
            # FIXME: need to find out why this happened #
            # log.error(error)
            pass

    def restart(self):
        pass



import os
import traceback

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType

from libs import logs
from libs.message import LogsCommand
from os import path
import requests
root_url = "http://logs-01.loggly.com/inputs/c58f8bb2-2332-4915-b9f3-70c1975956bb/tag/"
log = logs.get_logger(__name__)

class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    def handle_message(self, message):
        log.info('LogsManager handle message: %s' %(message.command_name))
        try:
            result = None
            if message.command_name == LogsCommand.send_logs:
                result = send_logs(message.content)
                type = ContentType.NONE

            # create reply message
            message.type = type
            message.content = result
            message.error = False
            self._send_to_node(message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, command_name=message.command_name)
            self._send_to_node(error_message)
       

def send_json_logs(json_data, tag):
    url = root_url + tag
    x = requests.post(url, json = json_data)
    log.info('Send log with tag: %s' %(tag) + " " + x.text)


def send_server_logs(tag):
    url = root_url + tag
    filepath = get_file_path("server.log")
    file =  open(filepath, "rb").read()
    x = requests.post(url, data = file)
    log.info('Send log with tag: %s' %(tag) + " " + x.text)


def get_file_path(filename):
    basepath = path.dirname(__file__)
    filepath = path.abspath(path.join(basepath, "..", "..", filename))
    return filepath


def send_workspace(tag):
    url = root_url + tag
    filepath = get_file_path("workspace.yaml")
    file =  open(filepath, "rb").read()
    x = requests.post(url, data = file)
    log.info('Send log with tag: %s' %(tag) + " " + x.text)


def send_logs(message):
    send_json_logs(message["clientLogs"], "clientLogs")
    send_json_logs(message["rootState"], "rootState")
    send_server_logs("serverLogs")
    send_workspace("workspace")
    return True



import time
import psutil
from traitlets.config import Configurable
from jupyter_resource_usage.prometheus import PrometheusHandler
from jupyter_resource_usage.metrics import PSUtilMetricsLoader
from jupyter_resource_usage.config import ResourceUseDisplay
import threading
import traceback
import simplejson as json

from libs.message_handler import BaseMessageHandler
from libs.message import Message

from libs import logs
from libs.message import WebappEndpoint
from libs.config import read_config
from libs.zmq_message import MessageQueuePull
from project_manager.interfaces import SERVER_CONFIG_PATH
from libs.message import ExecutorManagerCommand
from project_manager.interfaces import WORKSPACE_METADATA_PATH, WorkspaceMetadata
from server import set_executor_working_dir

log = logs.get_logger(__name__)

RESOUCE_CHECKING_PERIOD = 10  # unit: second
MEM_REPORT_THRESHOLD = 0.05

## This is fake server app used to make the jupyter_resource_usage plugin works #


class ServerApp(Configurable):
    class WebApp:
        settings = {}
    web_app = WebApp()
    session_manager = None


class ResourceStatus:
    alive_status: bool = False
    resource_usage = {"rss": 0., "limit": {}}


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)
        executor_manager_thread = threading.Thread(
            target=self.handle_message, args=(user_space,), daemon=True)
        executor_manager_thread.start()

        server_app = ServerApp()
        resuseconfig = ResourceUseDisplay(parent=server_app)
        server_app.web_app.settings["jupyter_resource_usage_display_config"] = resuseconfig
        self.config = resuseconfig
        self.metrics_manager = PrometheusHandler(
            PSUtilMetricsLoader(server_app))
        executor_status_thread = threading.Thread(
            target=self._update_resouce_usage, daemon=True)
        executor_status_thread.start()

        self.resource_status = ResourceStatus()

    def _check_resource_threshold(self, alive_status, resource_usage):
        if self.resource_status.alive_status != alive_status:
            self.resource_status.alive_status = alive_status
            return True
        else:
            if self.resource_status.resource_usage["rss"]==0. or ("rss" in resource_usage and abs(resource_usage["rss"]-self.resource_status.resource_usage["rss"])/self.resource_status.resource_usage["rss"] > MEM_REPORT_THRESHOLD):
                self.resource_status.resource_usage = resource_usage
                return True
        return False

    def _update_resouce_usage(self):
        message = Message(**{'webapp_endpoint': WebappEndpoint.ExecutorManager,
                             'command_name': ExecutorManagerCommand.get_status})
        while(True):
            # start_time = time.time()
            alive_status = self.user_space.is_alive()
            self.metrics_manager()
            resource_usage = self._get_resource_usage()
            # log.info('Resource: %s' % resource_usage)
            message.content = {'alive_status': alive_status,
                               'resource_usage': resource_usage}
            if self._check_resource_threshold(alive_status, resource_usage):
                self._send_to_node(message)
            # end_time = time.time()
            # log.info("Time %f" % (end_time-start_time))
            time.sleep(RESOUCE_CHECKING_PERIOD)

    def _get_cpu_percent(self, all_processes):
        def get_cpu_percent(p):
            try:
                return p.cpu_percent(interval=0.05)
            # Avoid littering logs with stack traces complaining
            # about dead processes having no CPU usage
            except:
                return 0

        return sum([get_cpu_percent(p) for p in all_processes])

    def _get_resource_usage(self):
        # self.metrics_manager()
        cur_process = psutil.Process()
        all_processes = [cur_process] + \
            cur_process.children(recursive=True)

        # Get memory information
        rss = 0
        for p in all_processes:
            try:
                rss += p.memory_info().rss
            except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
                pass

        if callable(self.config.mem_limit):
            mem_limit = self.config.mem_limit(rss=rss)
        else:  # mem_limit is an Int
            mem_limit = self.config.mem_limit

        limits = {"memory": {"rss": mem_limit}}
        if self.config.mem_limit and self.config.mem_warning_threshold != 0:
            limits["memory"]["warn"] = (mem_limit - rss) < (
                mem_limit * self.config.mem_warning_threshold
            )

        metrics = {"rss": rss, "limits": limits}

        # Optionally get CPU information
        if self.config.track_cpu_percent:
            cpu_count = psutil.cpu_count()
            cpu_percent = self._get_cpu_percent(all_processes)

            if self.config.cpu_limit != 0:
                limits["cpu"] = {"cpu": self.config.cpu_limit}
                if self.config.cpu_warning_threshold != 0:
                    limits["cpu"]["warn"] = (self.config.cpu_limit - cpu_percent) < (
                        self.config.cpu_limit * self.config.cpu_warning_threshold
                    )

            metrics.update(cpu_percent=cpu_percent, cpu_count=cpu_count)
        return metrics

    def handle_message(self, message):
        log.info("Kernel control thread started")
        ## reading config here to get the most updated version #
        server_config = read_config(SERVER_CONFIG_PATH)
        n2p_queue = MessageQueuePull(
            server_config.n2p_comm['host'], server_config.n2p_comm['kernel_control_port'])
        try:
            while True:
                strMessage = n2p_queue.receive_msg()
                message = Message(**json.loads(strMessage))
                log.info("Received control message: %s" % message)
                if message.command_name == ExecutorManagerCommand.restart_kernel:
                    result = self.user_space.restart_executor()
                    if result:
                        # get the lastest config to make sure that it is updated with the lastest open project
                        workspace_info = read_config(WORKSPACE_METADATA_PATH)
                        workspace_metadata = WorkspaceMetadata(
                            workspace_info.__dict__)
                        set_executor_working_dir(
                            self.user_space, workspace_metadata)
                    message = Message(**{'webapp_endpoint': WebappEndpoint.ExecutorManager,
                                         'command_name': message.command_name,
                                         'content': {'success': result}})
                elif message.command_name == ExecutorManagerCommand.interrupt_kernel:
                    result = self.user_space.interrupt_executor()
                    message = Message(**{'webapp_endpoint': WebappEndpoint.ExecutorManager,
                                         'command_name': message.command_name,
                                         'content': {'success': result}})
                elif message.command_name == ExecutorManagerCommand.get_status:
                    status = self.user_space.is_alive()
                    # resource_usage = self._get_resource_usage()
                    # log.info("Memory usage %s" % resource_usage)
                    message = Message(**{'webapp_endpoint': WebappEndpoint.ExecutorManager,
                                         'command_name': message.command_name,
                                         'content': {'alive': status}})
                    #  'content': {'alive': status, 'resource': resource_usage}})

                self._send_to_node(message)
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))
            error_message = BaseMessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name, {})
            self._send_to_node(error_message)



import sys
import traceback
import threading
import signal
import traceback
import sentry_sdk
from libs import logs
import logging
import pandas as pd
from libs import logs
import simplejson as json
from libs.config import read_config
from code_editor import code_editor as ce
from executor_manager import executor_manager as execm
from dataframe_manager import dataframe_manager as dm
from experiment_manager import experiment_manager as em
from model_manager import model_manager as mm
from cassist import cassist as ca
from file_explorer import file_explorer as fe
from file_manager import file_manager as fm
from logs_manager import logs_manager as lm
from jupyter_server_manager import jupyter_server_manager as jsm
from libs.zmq_message import MessageQueuePush, MessageQueuePull
from libs.message import Message, WebappEndpoint, ExecutorManagerCommand, ExecutorType
from libs.message_handler import BaseMessageHandler
from libs.constants import TrackingModelType
from project_manager.interfaces import SERVER_CONFIG_PATH, WORKSPACE_METADATA_PATH, WorkspaceMetadata
from user_space.user_space import IPythonUserSpace, BaseKernelUserSpace
import cnextlib.dataframe as cd


log = logs.get_logger(__name__)


class ShutdownSignalHandler:
    running = True

    def __init__(self, message_handler, user_space):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)
        self.user_space = user_space
        self.message_handler = message_handler

    def get_running_status(self):
        return self.running

    def exit_gracefully(self, *args):
        log.info('ShutdownSignalHandler {}'.format(args))

        if self.message_handler != None:
            for key, value in self.message_handler.items():
                log.info('Shutdown {}'.format(key))
                value.shutdown()
                # value.user_space.executor.interupt_kernel()

        self.running = False
        # currently we exit right here. In the future, consider option to stop message handler gracefully.
        sys.exit(0)


def get_active_project_path(workspace_info: WorkspaceMetadata):
    project_path = None
    if workspace_info.active_project is not None:
        for project in workspace_info.open_projects:
            if workspace_info.active_project == project.id:
                project_path = project.path
    return project_path


def set_executor_working_dir(user_space, workspace_info: WorkspaceMetadata):
    project_path = get_active_project_path(workspace_info)
    log.info('Project path: {}'.format(project_path))
    if project_path:
        user_space.set_executor_working_dir(project_path)


def main(argv):
    try:
        if argv and len(argv) > 0:
            server_config = read_config(SERVER_CONFIG_PATH)
            # workspace_metadata = read_config(WORKSPACE_METADATA_PATH)
            workspace_metadata = WorkspaceMetadata(
                read_config(WORKSPACE_METADATA_PATH).__dict__)

            executor_type = argv[0]
            user_space = None
            message_handler = None
            try:
                p2n_queue = MessageQueuePush(
                    server_config.p2n_comm['host'], server_config.p2n_comm['port'])
                jupyter_server_config = server_config.jupyter_server
                if executor_type == ExecutorType.CODE:
                    user_space = IPythonUserSpace(
                        (cd.DataFrame, pd.DataFrame), (TrackingModelType.PYTORCH_NN, TrackingModelType.TENSORFLOW_KERAS))

                    executor_manager = execm.MessageHandler(
                        p2n_queue, user_space)

                    message_handler = {
                        WebappEndpoint.CodeEditor: ce.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.DFManager: dm.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.ModelManager: mm.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.MagicCommandGen: ca.MessageHandler(
                            p2n_queue, user_space)
                    }

                    set_executor_working_dir(user_space, workspace_metadata)

                elif executor_type == ExecutorType.NONCODE:
                    user_space = BaseKernelUserSpace()
                    message_handler = {
                        WebappEndpoint.ExperimentManager: em.MessageHandler(p2n_queue, user_space),
                        WebappEndpoint.FileManager: fm.MessageHandler(p2n_queue, user_space, workspace_metadata),
                        WebappEndpoint.FileExplorer: fe.MessageHandler(
                            p2n_queue, user_space),
                        WebappEndpoint.Terminal: jsm.MessageHandler(p2n_queue, user_space, workspace_metadata, jupyter_server_config),
                        WebappEndpoint.LogsManager: lm.MessageHandler(p2n_queue, user_space),
                    }

            except Exception as error:
                log.error("%s - %s" % (error, traceback.format_exc()))
                exit(0)

            shutdowHandler = ShutdownSignalHandler(message_handler, user_space)
            # this condition here is meaningless for now because the process will be exit inside ShutdownSignalHandler.exit_gracefully already
            try:
                # while shutdowHandler.running:
                for line in sys.stdin:
                    try:
                        # log.info('Got message %s' % line)
                        message = Message(**json.loads(line))
                        log_content = ""
                        if isinstance(message.content, str):
                            if len(message.content) > 50:
                                log_content = message.content[:50]
                            else:
                                log_content = message.content
                        log.info('Got message from %s command: "%s", content: \'%s\'' %
                                 (message.webapp_endpoint, message.command_name, log_content))
                        message_handler[message.webapp_endpoint].handle_message(
                            message)
                    except OSError as error:  # TODO check if this has to do with buffer error
                        # since this error might be related to the pipe, we do not send this error to nodejs
                        log.error("OSError: %s" % (error))

                    except:
                        log.error("Failed to execute the command %s",
                                  traceback.format_exc())
                        message = BaseMessageHandler._create_error_message(
                            message.webapp_endpoint, traceback.format_exc(), message.command_name)
                        # send_to_node(message)
                        BaseMessageHandler.send_message(
                            p2n_queue, message.toJSON())

                    try:
                        sys.stdout.flush()
                    except Exception as error:
                        log.error("Failed to flush stdout %s - %s" %
                                  (error, traceback.format_exc()))
            except Exception as error:
                log.error("Exception %s - %s" %
                          (error, traceback.format_exc()))
                ## make sure all the message handler shut down properly #
                shutdowHandler.exit_gracefully()
                exit(0)

    except Exception as error:
        log.error("Failed to flush stdout %s - %s" %
                  (error, traceback.format_exc()))
        exit(0)


if __name__ == "__main__":

    # Initialize sentry
    # sentry_sdk.init(
    #     "https://318e8d1b2b0b43fbbbd38ea0426378b4@o1259763.ingest.sentry.io/6435268",

    #     # Set traces_sample_rate to 1.0 to capture 100%
    #     # of transactions for performance monitoring.
    #     # We recommend adjusting this value in production.
    #     traces_sample_rate=0.1
    # )

    main(sys.argv[1:])



import glob
import os
from site import abs_paths
import sys
from subprocess import Popen
from contextlib import contextmanager
import time

from urllib.request import urlopen
from io import BytesIO
from zipfile import ZipFile
import uuid
import shutil
import yaml
import os
from os import path as _path
from pathlib import Path
if sys.platform.startswith("win"):
    from pyreadline import Readline
    readline = Readline()
else:
    import readline


current_dir_path = os.getcwd()
basepath = _path.dirname(__file__)
path = Path(basepath)
current_dir_path = path.absolute()


PROJECTS_PATH = os.path.abspath(os.path.join(current_dir_path,"projects"))
SERVER_PATH = os.path.abspath(os.path.join(current_dir_path,"server"))
DEFAULT_PROJECT = "Skywalker"
WITHOUT_PROJECT = 0
HAVE_PROJECT = 1
DOWNLOAD_PATH = 'https://bitbucket.org/robotdreamers/cnext_sample_projects/get/master.zip'

def change_permissions_recursive(path, mode):
    for root, dirs, files in os.walk(path, topdown=False):
        for dir in [os.path.join(root, d) for d in dirs]:
            os.chmod(dir, mode)
    for file in [os.path.join(root, f) for f in files]:
        os.chmod(file, mode)


def change_workspace(name, path):
    project_id = str(uuid.uuid1())
    data = {
        'active_project': project_id,
        'open_projects': [{"id": project_id, 'name': name, 'path': path}],
    }
    with open(r'cnext_server/workspace.yaml', 'w') as file:
        documents = yaml.dump(data, file, default_flow_style=False)
    print('change workspace done!')


def install_node_modules():
    os.chdir(SERVER_PATH)
    os.system('npm i')

samples = {
    "skywalker": "Skywalker",
    "jedi": "Jedi",
}

def download_and_unzip(url, project_name, extract_to='.'):
    if not os.path.exists(PROJECTS_PATH): 
        os.makedirs(PROJECTS_PATH)

    http_response = urlopen(url)
    zipfile = ZipFile(BytesIO(http_response.read()))
    zipfile.extractall(path=extract_to)

    # grand per
    change_permissions_recursive(extract_to, 0o777)

    # cut - paste
    shutil.copytree(src=os.path.join(extract_to, zipfile.namelist()[
                    0], project_name), dst=os.path.join(extract_to, project_name), dirs_exist_ok=True)
    
    # remove old folder
    shutil.rmtree(path=os.path.join(extract_to, zipfile.namelist()[0]))
  


def complete(text, state):
    return (glob.glob(os.path.expanduser(text+'*'))+[None])[state]


readline.set_completer_delims(' \t\n;')
readline.parse_and_bind("tab: complete")
readline.set_completer(complete)

def main(args=sys.argv):
        if(len(args) == 1):
            start()
        elif(len(args) == 2):
            if(args[1].isnumeric()):
                run_at_port(int(args[1]))
            else:
                switch(args[1])
        elif(len(args) == 3):
            switch(args[1],args[2])
        else:
            default()

def run_help(choice):
    message = """
        Installation command
        - cnext -s                     : START with `Skywalker` project
        - cnext -s Jedi                : START with `Jedi` project
        - cnext -s G:\DEV\PROJECTS     : START with `Skywalker` inside PROJECTS

        Using command
        - cnext                        : RESUME APPLICATION 
        - cnext 8888                   : RESUME APPLICATION at PORT 8888
        """
    print(message)
    
def default(data):
    run_help(data)
    
def run_at_port(port):
    start(port)
   

def download_project(project_name, download_to_path):
    download_and_unzip(DOWNLOAD_PATH, project_name, download_to_path)
    project_path = os.path.join(download_to_path, project_name)
    change_workspace(project_name, os.path.normpath(project_path).replace(os.sep, '/'))


def start_with_sample_project(path_or_name):
    
    if(path_or_name):
        abs_paths = os.path.abspath(path_or_name)
        if os.path.isdir(abs_paths):
            # cnext -s G:\DEV\PROJECTS
            download_project(DEFAULT_PROJECT,abs_paths)
        elif(path_or_name.lower() in list(samples.keys())):
            # cnext -s skywalker|Jedi
            download_project(samples[path_or_name.lower()], PROJECTS_PATH)
        else:
            print("your path or name isn't correctly")
            return
    else:
        # cnext -s == cnext -s skywalker
        download_project(DEFAULT_PROJECT,PROJECTS_PATH)
    
    install_node_modules()
    start()

switcher = {
    "-h": run_help, 
    "-s": start_with_sample_project, 
    # full case
    "--help": run_help,
    "--start": start_with_sample_project,
}

def switch(command, data = None ):
    return switcher.get(command, default)(data)

@contextmanager
def run_and_terminate_process(port):
    try:
        os.chdir(SERVER_PATH)
        my_env = os.environ.copy()
        my_env["PATH"] = os.path.dirname(
            sys.executable) + os.path.pathsep + my_env["PATH"]
        command = "set PORT="+ f'{port}' + "&& node server.js"
        ser_proc = Popen(command, shell=True, env=my_env)
        yield

    finally:
        ser_proc.terminate()  # send sigterm, or ...
        ser_proc.kill()      # send sigkill


def start(port=4000):
    with run_and_terminate_process(port) as running_proc:
        while True:
            time.sleep(1000)


if __name__ == '__main__':
    start()
